#!/bin/bash

# Set global variables:
#   - servers
#   - redundant_conf 
#   - valid_redundant_conf
#   - remote_server_indeces
#   - my_ip_addr
#   - my_id
#   - pm
#   - el
#   - fifo_dir
#   - tail_filter
#   - logfile
#
# Fix the owner/group setting of /opt/passlogic/data/conf/redundant.conf
# to passlogic:passlogic 
# (restore.sh modifies it to apache:apache.)
#
pm_init() {
    local ip_com line i

    redundant_conf=/opt/passlogic/data/conf/redundant.conf
    ip_com=`/sbin/ip -f inet addr`

    [ ! -e $redundant_conf ] && touch $redundant_conf
    chown passlogic:passlogic $redundant_conf

    servers=()
    for line in `sed -e 's/SERVER[0123]=//' $redundant_conf`; do
	servers=("${servers[@]}" $line)
	echo $ip_com | grep -q "${line}/" && my_ip_addr=$line
    done

    [ -z $my_ip_addr ] && my_ip_addr='localhost'

    [ ${#servers[@]} -lt 2 ] && valid_redundant_conf=false || valid_redundant_conf=true

    remote_server_indeces=()
    for i in $(seq 0 $((${#servers[@]}-1))); do
	if [ ! $my_ip_addr = ${servers[$i]} ]
	then
	    remote_server_indeces=("${remote_server_indeces[@]}" $i)
	else
	    my_id=$i
	fi
    done

    pm=/opt/passlogic/apps/tools/pm.sh
    el=`sed 's/.*release\s\([0-9]\).*/\1/' /etc/redhat-release`
    fifo_dir=/home/passlogic

    if [ -z "$verbose" ]
    then
	tail_filter="tail -1"
	exec 2>/dev/null
    else
	tail_filter="cat"
    fi

    logfile=/var/log/passlogic/pm.log
}

pipestatus() {
    local ps=("${PIPESTATUS[@]}")
    local argv=("$@")
    local ret=0
    local i

    for ((i=0; i<$#; ++i))
    do
	[ ! ${ps[$i]} -eq ${argv[$i]} ] && ret=1 && break
    done
    return $ret
}

show_activity() { echo "`LANG=C date "+%Y-%m-%d %H:%M:%S"`: $my_ip_addr: $*" | tee -a $logfile ; }
log_activity() { echo "`LANG=C date "+%Y-%m-%d %H:%M:%S"`: $my_ip_addr: $*" >> $logfile ; }

show_ok_ng() { [ $? -eq 0 ] && echo :ok || echo :ng; }

service_() {
    case $el in
	"6")
	    /sbin/service $1 $2
	    ;;
	"7")
	    /bin/systemctl $2 $1
	    ;;
    esac
}

# name of PostgreSQL process in an output of 'ps -upasslogic -o comm'
#
postmaster() {
    case $el in
	"6")
	    echo postmaster
	    ;;
	"7")
	    echo postgres
	    ;;
    esac
}

server_config() {
	/opt/passlogic/pgsql/bin/psql -h localhost -p 5439 -U passlogic -d passlogic -t -c "select config from passlogic_config where type='server_config'"
}

read_server_config() {
	echo `server_config` | sed -e "s/^.*\"$1\":\"\([0-9]\+\)\",.*$/\1/" | head -1
}


wait_socket_5439() { while : ; do netstat -tnlp 2> /dev/null | grep `postmaster` | grep -q 5439 && break; done; echo :ok;}

wait_pgsql_up() {
    while :
    do
	/opt/passlogic/pgsql/bin/psql -h localhost -p 5439 -d passlogic -U passlogic -t -c "select 'ready'" | grep -q "ready" && break
    done
    echo :ok
}

wait_pgpool_up() {
    while :
    do
	/opt/passlogic/pgsql/bin/psql -h localhost -p 9915 -d passlogic -U passlogic -t -c "select 'ready'" | grep -q "ready" && break
    done
    echo :ok
}

wait_socket_9915_9925() {
    while :
    do
	netstat -tnlp 2>/dev/null | grep pgpool | grep -q 9915 &&
	    netstat -tnlp 2>/dev/null | grep pgpool | grep -q 9925 && break
    done
    echo :ok
}

wait_postmaster_disappear() { while : ; do ps -upasslogic -o comm | grep -q -E `postmaster` || break; done; echo :ok; }

wait_pgpool_disappear() { while : ; do ps -upasslogic -o comm | grep -q pgpool || break; done; echo :ok; }

wait_httpd_disappear() { while : ; do ps -e -o comm | grep -q -E '^httpd$' || break; done; echo :ok; }

ping() {
    echo "pong from $my_ip_addr"
    echo :ok
}

pgsql_down() {
    service_ passlogic-pgsql stop
    timeout 10 $pm <<< "wait_postmaster_disappear"
    [ $? -eq 124 ] && echo :ng || echo :ok
}

pgsql_up_check() {
    local ret
    timeout 10 $pm <<< "wait_socket_5439"
    if [ $? -eq 124 ]
    then
	ret=:ng
    else
	timeout 10 $pm <<< "wait_pgsql_up"
	[ $? -eq 124 ] && ret=:ng || ret=:ok
    fi
    echo $ret
}

pgsql_up() { service_ passlogic-pgsql start; show_ok_ng; }

pgsql_status() {
    timeout 10 /opt/passlogic/pgsql/bin/psql -p 5439 -d passlogic -U passlogic -t -c "select 'ready'" | grep -q "ready"
    pipestatus 0 0 && echo :ok || echo :ng
}

pgpool_down() {
    service_ passlogic-pgpool stop
    timeout 10 $pm <<< "wait_pgpool_disappear"
    [ $? -eq 124 ] && echo :ng || echo :ok
}

pgpool_up() {
    local ret
    service_ passlogic-pgpool start
    timeout 10 $pm <<< "wait_socket_9915_9925"
    if [ $? -eq 124 ]
    then
	ret=:ng
    else
	timeout 10 $pm <<< "wait_pgpool_up"
	[ $? -eq 124 ] && ret=:ng || ret=:ok
    fi
    echo $ret
}

pgpool_status() {
    timeout 10 /opt/passlogic/pgsql/bin/psql -p 9915 -d passlogic -U passlogic -t -c "select 'ready'" | grep -q "ready"
    pipestatus 0 0 && echo :ok || echo :ng
}

node_info() {
    local node_info
    node_info=`PCPPASSFILE=/opt/passlogic/lib/pcppass.conf /opt/passlogic/pgpool/bin/pcp_node_info -h localhost -p 9925 -U passlogic -w -n $1`
    if [ $? -eq 0 ]
    then
	echo $node_info | sed 's/.*\s5439\s\([0-9]\).*/\1/'
    else
	echo :ng
    fi
}

detach_node() {
    PCPPASSFILE=/opt/passlogic/lib/pcppass.conf /opt/passlogic/pgpool/bin/pcp_detach_node -h localhost -p 9925 -U passlogic -w -n $1
}

attach_node() {
    PCPPASSFILE=/opt/passlogic/lib/pcppass.conf /opt/passlogic/pgpool/bin/pcp_attach_node -h localhost -p 9925 -U passlogic -w -n $1
}

detach_remote() {
    local status ret
    ret=:ok
    for i in ${remote_server_indeces[@]}
    do
	[ `node_info $i` = "3" ] || detach_node $i
	[ `node_info $i` = ":ng" ] && ret=:ng && break
    done
    echo $ret
}

attach_remote() {
    local status ret
    ret=:ok
    for i in ${remote_server_indeces[@]}
    do
	attach_node $i
	grep -q -E '^[12]' <<< `node_info $i`
	[ ! $? -eq 0 ] && ret=:ng && break
    done
    echo $ret
}

httpd_down() {
    service_ httpd stop
    timeout 10 $pm <<< "wait_httpd_disappear"
    [ $? -eq 124 ] && echo :ng || echo :ok
}

httpd_up() { service_ httpd start; show_ok_ng; }

httpd_graceful() {
    local com
    case $el in
	"6")
	    com=graceful
	    ;;
	"7")
	    com=reload
	    ;;
    esac
    service_ httpd $com
    show_ok_ng;
}

httpd_status() { service_ httpd status; show_ok_ng; }

radiusd_restart() { service_ radiusd restart; show_ok_ng; }

# Copy following directories;
#   - /opt/passlogic/pgsql/data
#
# which contains files generated by passlogic_redundant.sh
#
#   - /opt/passlogic/pgsql/data/pg_hba.conf
#   - /opt/passlogic/pgsql/data/failover_mail.sh
#
copy_pgsql_archives_data_dirs() {
    local i rsync_com
    for i in ${remote_server_indeces[@]}
    do
	echo copying to ${servers[$i]} ...
	rsync_com=$(cat << EOS
        rsync -v -c -az --delete -e ssh \
              --exclude postmaster.opts \
              --exclude postmaster.pid \
              --exclude serverlog \
              --exclude bin \
              --exclude include \
              --exclude lib \
              --exclude share \
              /opt/passlogic/pgsql/ ${servers[$i]}:/opt/passlogic/pgsql/
EOS
)
	echo $rsync_com
	su - passlogic -c "$rsync_com"
	show_ok_ng
    done
}

# Copy following files generated by passlogic_redundant.sh
#
#   - /opt/passlogic/pgpool/etc/pgpool.conf
#   - /opt/passlogic/data/conf/redundant.conf
#
copy_pgpool_redundant_conf() {
    local i scp_com
    for i in ${remote_server_indeces[@]}
    do
	scp_com=$(cat << EOS
        scp -p /opt/passlogic/pgpool/etc/pgpool.conf ${servers[$i]}:/opt/passlogic/pgpool/etc/ &&
        scp -p /opt/passlogic/data/conf/redundant.conf ${servers[$i]}:/opt/passlogic/data/conf/
EOS
)
	echo $scp_com
	su - passlogic -c "$scp_com"
	show_ok_ng
    done
}

remove_cache() { rm -f /opt/passlogic/tmp/*.cache; show_ok_ng; }

truncate_log_file() { truncate /var/log/passlogic/$1 --size 0; show_ok_ng; }

rotate_log_file() {
	chown apache:apache /var/log/passlogic/$1 &&
	cp /var/log/passlogic/$1 /var/log/passlogic/$1-99999999 &&
	chown apache:apache /var/log/passlogic/$1-99999999 &&
	truncate /var/log/passlogic/$1 --size 0
	show_ok_ng
}

flock_rotate_log_file() {
	flock -e /var/log/passlogic/$1 $pm <<< "rotate_log_file $1" | tail -1
}

swap_log_file() {
	chown apache:apache /var/log/passlogic/$1 &&
	cp /var/log/passlogic/$1 /var/log/passlogic/.$1 &&
	truncate /var/log/passlogic/$1 --size 0 &&
	cat /var/log/passlogic/$1-99999999 >> /var/log/passlogic/$1 &&
	cat /var/log/passlogic/.$1 >> /var/log/passlogic/$1 &&
	rm -f /var/log/passlogic/$1-99999999 /var/log/passlogic/.$1
	show_ok_ng
}

flock_swap_log_file() {
	flock -e /var/log/passlogic/$1 $pm <<< "swap_log_file $1" | tail -1
}

create_redundant_files() {
	local flag_dir=/opt/passlogic/data/conf/flag

	touch ${flag_dir}/redundant &&
	touch ${flag_dir}/redundant_status &&
	echo -n ${servers[0]} > ${flag_dir}/server_0 &&
	echo -n ${servers[1]} > ${flag_dir}/server_1 &&
	echo -n server_${my_id} > ${flag_dir}/my_server &&
	echo -n server_${remote_server_indeces} >${flag_dir}/other_server &&
	chown apache:apache ${flag_dir}/redundant ${flag_dir}/server_{0,1} ${flag_dir}/{my,other}_server &&
	chown passlogic:passlogic ${flag_dir}/redundant_status
	show_ok_ng
}

create_redundant_status() {
	create_status redundant
}

remove_redundant_status() {
	remove_status redundant
}

create_need_recovery_status() {
	create_status need_recovery
}

remove_need_recovery_status() {
	remove_status need_recovery
}

create_status() {
	local status=/opt/passlogic/data/conf/flag/$1_status

	if $valid_redundant_conf
	then
		touch ${status} && chown passlogic:passlogic ${status}
	fi
	show_ok_ng
}

remove_status() {
	local status=/opt/passlogic/data/conf/flag/$1_status

	rm -f  ${status}
	show_ok_ng
}

main_server_up() {
	touch /opt/passlogic/data/conf/flag/main_server && \
		chown apache:apache /opt/passlogic/data/conf/flag/main_server
	show_ok_ng
}
main_server_down() { rm -f /opt/passlogic/data/conf/flag/main_server; show_ok_ng; }

reloadCronsetting() { unset TERM; /opt/passlogic/apps/tools/reloadCronsetting.php; show_ok_ng; }
reloadPKIsetting() { unset TERM; /opt/passlogic/apps/tools/reloadPKIsetting.php; show_ok_ng; }
reloadProxysetting() { unset TERM; /opt/passlogic/apps/tools/reloadProxysetting.php; show_ok_ng; }
reloadRadiussetting() { unset TERM; /opt/passlogic/apps/tools/reloadRadiussetting.php; show_ok_ng; }
recreate_setting() { unset TERM; /opt/passlogic/apps/tools/recreateSetting.php; show_ok_ng; }
reloadTenant() { unset TERM; /opt/passlogic/apps/tools/kick_reloadTenant.php; show_ok_ng; }

cleanup_session_table() {
    local ret lifetime cond_t

    lifetime=`read_server_config PL_SESSION_TIMEOUT`
    grep -q -E '^[0-9]+$' <<< "$lifetime"
    if [ $? -eq 0 ]
    then
	cond_t=$(( `date +%s` - $lifetime ))
	/opt/passlogic/pgsql/bin/psql -h localhost -p 5439 -d passlogic -U passlogic -t <<EOF
	delete from passlogic_session where updated_at < $cond_t
EOF
	[ $? -eq 0 ] && ret=:ok || ret=:ng
    else
	ret=:ng
    fi
    echo $ret
}

transfer_settings() {
    local i rsync_com
    for i in ${remote_server_indeces[@]}
    do
	echo copying to ${servers[$i]} ...
	rsync_com=$(cat << EOS
        rsync -a -e ssh --delete /home/passlogic/exported_settings/${my_id} ${servers[$i]}:/home/passlogic/exported_settings/
EOS
)
	echo $rsync_com
	su - passlogic -c "$rsync_com"
	show_ok_ng
    done
}

export_settings() {
    su - passlogic -c "mkdir -m 755 -p /home/passlogic/exported_settings/"
    su - passlogic -c "mkdir -m 755 -p /home/passlogic/exported_settings/${my_id}"
    rsync -a --exclude redundant.conf --delete /opt/passlogic/data/ /home/passlogic/exported_settings/$my_id/data/ && \
	chown -R passlogic:passlogic  /home/passlogic/exported_settings/
    show_ok_ng
}

import_conf() {
    fsync logo_images.png apache:apache $1
    show_ok_ng
}

import_license() {
    local master_node_id=$1
    cp -f --preserve=timestamps /home/passlogic/exported_settings/${master_node_id}/data/license-ent.asc /opt/passlogic/data/ && \
	chown apache:apache /opt/passlogic/data/license-ent.asc
    show_ok_ng
}

lock() {
	local pid=$1
	local lockfile_basename=$2
	local lockfile=/opt/passlogic/data/lock/$2
	local interval=$3
	local process_monitor=${@:4}
	local ret
	
	mkdir -p /opt/passlogic/data/lock
	chown passlogic:passlogic /opt/passlogic/data/lock
	chmod 755 /opt/passlogic/data/lock
	su - passlogic -c "flock $lockfile bash -c \"echo -n $pid > $lockfile && chmod 444 $lockfile\""
	ret=$?

	nohup $pm <<< "lock_col $pid $lockfile_basename $interval $process_monitor" > /dev/null 2>&1 &

	[ $ret -eq 0 ]
	show_ok_ng
}

unlock() {
	local pid=$1
	local lockfile=/opt/passlogic/data/lock/$2
	local ret=0
	
	if [ -r $lockfile ] && [ `cat $lockfile` == "$pid" ]
	then
		rm $lockfile
		ret=0
	else
		ret=1
	fi

	[ $ret -eq 0 ]
	show_ok_ng
}

lock_col() {
	local pid=$1
	local lockfile=/opt//passlogic/data/lock/$2
	local lockfile_basename=$2
	local interval=$3
	local process_monitor=${@:4}
	local ret=0

	# PIPE -> KILL を止める
	trap '' SIGPIPE
	# TERM -> ロックファイル削除
	trap 'rm -f $lockfile; log_activity lock_col receieved SIGTERM. lockfile removed' SIGTERM
	while :
	do
		[ ! -r $lockfile ] && break
		[ `cat $lockfile` != "$pid" ] && ret=1 && break
		$process_monitor
		if [ $? -eq 1 ]
		then
			if [ -r $lockfile  ]
			then
				log_activity lock_col unlockd $lockfile_basename
				$pm <<< "unlock $pid $lockfile_basename"
				break
			fi
		fi
		sleep $interval
	done
	[ $ret -eq 0 ]
	show_ok_ng
}

# proc_mon
#
#   引数を=で区切って
#   １番目をps, ２番目をgrep の引数として扱う
# 
proc_mon() {
	local ps_opt
	local grep_opt

	ps_opt=`echo $* | cut -f 1 -d'='`
	grep_opt=`echo $* | cut -f 2 -d'='`

	ps ${ps_opt} | grep -v grep | grep -q ${grep_opt}
	show_ok_ng
}

replication?() {
	$valid_redundant_conf
	show_ok_ng
}

# o sync a file in exported_settings/${master_node_id}/data/conf/ with /opt/passlogic/data/conf/
# o change a file owner/group to a specified owner/group
fsync() {
    local file=$1
    local owner_group=$2
	local master_node_id=$3
    ( ( [ ! -e /home/passlogic/exported_settings/${master_node_id}/data/${file} ] && rm -f /opt/passlogic/data/${file} ) || \
      ( cp -f --preserve=timestamps /home/passlogic/exported_settings/${master_node_id}/data/${file} /opt/passlogic/data/${file} && \
	      chown ${owner_group} /opt/passlogic/data/${file}) )
}

send() {
    local argv=("$@")
    local host=${servers[${argv[0]}]}
    echo "${argv[@]:1}" | su - passlogic -c "ssh -T $host sudo $pm"
}

call_internal_func() {
    $* 2>&1 | tee -a $logfile | tail -1 | grep -q -E '^:ok$'; show_ok_ng
}

call_remote_func() {
    local line
    for i in ${remote_server_indeces[@]}; do
	send $i $* 2>&1 | tee -a $logfile | while read line
	do
	    grep -q -E '^:ok$' <<< "$line"
	done
	show_ok_ng
    done
}

show_error_message() { echo $* | tee -a $logfile; echo :ng | tee -a $logfile; }

while getopts "vs" OPT
do
    case $OPT in
	v)
	    verbose=0
	    ;;
	s)
		stream='yes'
		;;
    esac
done

shift $(($OPTIND-1))

pm_init

if [ -n "$stream" ]
then
	ary=(`echo $* | tr -s ' '`)
	c=${ary[0]}
    opts=("${ary[@]:1}")
    log_activity $c "${opts[@]}"

	case $c in
		# send_files receiver_dir sender_dir filenames
		#   sender_dir filenames => rest_opts
		#   ex) send_files /var/log/passlogic/ /var/log/passlogic/ `find /var/log/passlogic -name 'passlogic.log*' -printf '%f '`
		#
		send_files)
			receiver_dir=${opts[0]}
			rest_opts=("${opts[@]:1}")
			host=${servers[${remote_server_indeces}]}
			tar zcpf - -C "${rest_opts[@]}" | su - passlogic -c "ssh -T $host sudo $pm -s receive_files $receiver_dir"
			;;
		receive_files)
			tar zxpf - -C "${opts[@]}"
			;;
	esac
	ret=$?
	log_activity $c "${opts[@]}" terminated with code $ret
	exit $?
fi

while read msg
do
    ary=(`echo $msg | tr -s ' '`)
    c=${ary[0]}
    opts=("${ary[@]:1}")
    show_activity $c "${opts[@]}"
    case $c in
	ping|pgsql_*|pgpool_*|httpd_*|radiusd_restart|remove_cache|truncate_log_file|rotate_log_file|flock_rotate_log_file|swap_log_file|flock_swap_log_file|recreate_setting|reloadTenant|cleanup_session_table|create_redundant_status|create_need_recovery_status|wait_*|lock|lock_col|proc_mon|unlock|log_activity|replication?)
	    call_internal_func $c "${opts[@]}"
	    ;;
	copy_*|export_settings|transfer_settings|import_*|??tach_remote|main_server_*|reload*setting|create_redundant_files|remove_redundant_status|remove_need_recovery_status|wait_*)
	    if $valid_redundant_conf
	    then
		call_internal_func $c "${opts[@]}"
	    else
		show_error_message "Error in calling $c: please confirm the servers configuration in $redundant_conf"
	    fi
	    ;;
	remote_*)
	    if $valid_redundant_conf
	    then
		remote_com=${c/remote_/}
		call_remote_func $remote_com "${opts[@]}"
	    else
		show_error_message "Error in calling $c: please confirm the servers configuration in $redundant_conf"
	    fi
	    ;;
	deploy_remote_conf|deploy_remote_license)
	    if $valid_redundant_conf
	    then
		remote_com=import_${c/deploy_remote_/}
		call_remote_func $remote_com $my_id
	    else
		show_error_message "Error in calling $c: please confirm the servers configuration in $redundant_conf"
	    fi
	    ;;
    esac
done | $tail_filter
